#include <GL/glut.h>
#include <iostream>
#include <ctime>
#include "model.h"
#include "framebuffer.h"
using namespace std;

#define PI 3.14159265

Model** modelPtr;

/* frame size */
int screenWidth = 800;
int screenHeight = 600;
int screenWidth_half = 400;
int screenHeight_half = 300;

/* theta */
double rotateSpeed = 0.05;
double thetaX = 0;
double thetaY = 0;

int curModelIdx;
bool culling;
Framebuffer framebuffer(screenWidth, screenHeight);

/* model names */
char* modelNames[] = {
	"model/quad.obj",
	"model/couch.obj",
	"model/blaze.obj",
	"model/ateneal.obj",
	"model/venusm.obj",
	"model/bunnyC.obj",
	"model/duck4KN.obj",
	"model/happy10KN.obj",
	"model/dragon10KN.obj",
	"model/elephant16KN.obj",
	"model/Statue_of_Liberty.obj",
	"model/Nissan_Pathfinder.obj"
};
const int numModels = sizeof(modelNames) / sizeof(char*);

void printHelp() 
{
	printf("===========================================\n");
	printf("  H/h: Show help menu                      \n");
	printf("  M/m: Select model                        \n");
	printf("  UP/DOWN: Rotate along x-axis             \n");
	printf("  LEFT/RIGHT: Rotate along y-axis          \n");
	printf("  C/c: Toggle back-face culling            \n");
	printf("  B/b: Toggle background color             \n");
	printf("  S/s: Save image                          \n");
	printf("  Q/q: Quit                                \n");
	printf("===========================================\n\n");
}

void init() 
{
	modelPtr = new Model*[numModels];
	for(int i = 0; i < numModels; ++i) {
		/* load models */
		modelPtr[i] = readObj(modelNames[i]);

		/* normalize coordinates to [-0.5, +0.5] */
		float xmax, xmin, ymax, ymin, zmax, zmin, longest;
		float xmean, ymean, zmean;
		xmax = xmin = modelPtr[i]->vertices[3];
		ymax = ymin = modelPtr[i]->vertices[4];
		zmax = zmin = modelPtr[i]->vertices[5];
		for(int v = 2; v <= modelPtr[i]->numVertices; ++v) {
			xmax = max(xmax, modelPtr[i]->vertices[3*v]);
			xmin = min(xmin, modelPtr[i]->vertices[3*v]);
			ymax = max(ymax, modelPtr[i]->vertices[3*v+1]);
			ymin = min(ymin, modelPtr[i]->vertices[3*v+1]);
			zmax = max(zmax, modelPtr[i]->vertices[3*v+2]);
			zmin = min(zmin, modelPtr[i]->vertices[3*v+2]);
		}
		longest = max(xmax - xmin, ymax - ymin);
		longest = max(longest, zmax - zmin);
		xmean = 0.5f*(xmax + xmin);
		ymean = 0.5f*(ymax + ymin);
		zmean = 0.5f*(zmax + zmin);
		
		for(int v = 1; v <= modelPtr[i]->numVertices; ++v) {
			modelPtr[i]->vertices[3*v  ] = (modelPtr[i]->vertices[3*v  ] - xmean)/longest;
			modelPtr[i]->vertices[3*v+1] = (modelPtr[i]->vertices[3*v+1] - ymean)/longest;
			modelPtr[i]->vertices[3*v+2] = (modelPtr[i]->vertices[3*v+2] - zmean)/longest;
		}
	}

	/* initialize parameters */
	curModelIdx = 0;
	culling = true;
}

void displayFunc() 
{
	framebuffer.clear();

	/* draw point cloud */
	/* TODO: change the following code to do rasterization */
	int ix, iy;
	int screenSide = min(screenWidth, screenHeight);
	float margin = 0.1f;
	float prevX, prevY, prevZ, curX, curY, curZ;
	for(int i = 1; i <= modelPtr[curModelIdx]->numVertices; ++i) {
		prevX = modelPtr[curModelIdx]->vertices[3*i  ];
		prevY = modelPtr[curModelIdx]->vertices[3*i+1];
		prevZ = modelPtr[curModelIdx]->vertices[3*i+2];

		// rotate along x-axis
		curY = prevY*cos(thetaX)-prevZ*sin(thetaX);
		curZ = prevY*sin(thetaX)+prevZ*cos(thetaX);
		// prevY = curY;
		prevZ = curZ;
		// rotate along y-axis
		curX = prevZ*sin(thetaY)+prevX*cos(thetaY);
		curZ = prevZ*cos(thetaY)-prevX*sin(thetaY);
		// draw the framebuffer
		ix = int(((1.f-margin)*curX)*screenSide) + screenWidth_half - 1;
		iy = int(((1.f-margin)*curY)*screenSide) + screenHeight_half - 1;
		framebuffer.draw(ix, iy, curZ, vec3(1.f));
	}

    /* display */
	glDrawPixels(screenWidth, screenHeight, GL_RGB, GL_FLOAT, (const GLvoid*)framebuffer.getPixels());
	glutSwapBuffers();

	/* FPS counter */
	static clock_t prev = clock();
	static clock_t curr;
	static char title[20];
	static double refreshTime = 0.5;
	static int count = 0;
	++count;
	curr = clock();
	double t = (double)(curr - prev)/(double)CLOCKS_PER_SEC;
	if (t > refreshTime) {
		sprintf(title, "3DMM HW#1 Rasterization: %.2lf fps",  (double)count/t);
		glutSetWindowTitle(title);
		prev = curr;
		count = 0;
	}
}


void idleFunc() 
{
	glutPostRedisplay();
}

void keyboardFunc(unsigned char key, int x, int y) 
{
	switch (key) {
	// Quit
	case 'q':
	case 'Q':
		exit(0);
		break;
	// Help
	case 'h':
	case 'H':
		printHelp();
		break;
	// Save image
	case 's':
	case 'S':
		static time_t t;
		static char name[80];
		time(&t);
		strftime(name, sizeof(name), "%Y%m%d%H%M%S.ppm", localtime(&t));
		printf("Save framebuffer to %s\n", name);
		framebuffer.writePPM(name);
		break;
	// Select model
	case 'm':
		curModelIdx = (curModelIdx == numModels - 1)? 0 : (curModelIdx + 1);
		printf("Switch to model \"%s\"\n", modelNames[curModelIdx]);
		break;
	case 'M':
		curModelIdx = (curModelIdx == 0)? (numModels - 1) : (curModelIdx - 1);
		printf("Switch to model \"%s\"\n", modelNames[curModelIdx]);
		break;
	// Back-face culling
	case 'c':
	case 'C':
		culling = !culling;
		break;
	// Background color
	case 'b':
	case 'B':
		static bool isBlack = true;
		isBlack = !isBlack;
		framebuffer.setClearColor(isBlack? vec3(0.f) : vec3(1.f));
		break;
	// You can add more functions as you like.
	}
	glutPostRedisplay();
}

// Special keys
void specialFunc(int key, int x, int y) 
{
	switch (key) {
	case GLUT_KEY_RIGHT:
		thetaY = (thetaY > PI*2)? 0 : (thetaY + rotateSpeed);
		break;
	case GLUT_KEY_LEFT:
		thetaY = (thetaY < 0)? (PI*2) : (thetaY - rotateSpeed);
		break;
	case GLUT_KEY_UP:
		thetaX = (thetaX < 0)? (PI*2) : (thetaX - rotateSpeed);
		break;
	case GLUT_KEY_DOWN:
		thetaX = (thetaX > PI*2)? 0 : (thetaX + rotateSpeed);
		break;
	case GLUT_KEY_PAGE_UP:
    
		break;
	case GLUT_KEY_PAGE_DOWN:
    
		break;
	}
}

int main(int argc, char** argv)
{
	glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGBA);
	glutInitWindowSize(screenWidth, screenHeight);
	glutCreateWindow("3DMM HW#1 Rasterization");

	init();

	glutDisplayFunc(displayFunc);
	glutIdleFunc(idleFunc);
	// glutMouseFunc(mouseFunc);
	glutKeyboardFunc(keyboardFunc);
	glutSpecialFunc(specialFunc);

	glutMainLoop();
	
	for(int i = 0; i < numModels; ++i) {
		if(modelPtr[i]) delete [] modelPtr[i];
	}
	if(modelPtr) delete [] modelPtr;

	return 0;
}